from dataclasses import dataclass
import datetime
import pathlib
from beartype.roar import BeartypeCallHintViolation
from typing import (
    Union,
    Any,
)

import pytest

import serde
import serde.json

from . import data

test_cases: list[tuple[Any, Any, bool]] = [
    (int, 10, False),
    (int, 10.0, True),
    (int, "10", True),
    (int, True, False),
    (float, 10, True),
    (float, 10.0, False),
    (float, "10", True),
    (float, True, True),
    (str, 10, True),
    (str, 10.0, True),
    (str, "10", False),
    (str, True, True),
    (bool, 10, True),
    (bool, 10.0, True),
    (bool, "10", True),
    (bool, True, False),
    (list[int], [1], False),
    (list[int], [1.0], True),
    (list[float], [1.0], False),
    (list[float], [1], True),
    (list[float], ["foo"], True),
    (list[str], ["foo"], False),
    (list[str], [True], True),
    (list[bool], [True], False),
    (list[bool], [10], True),
    (list[data.Int], [data.Int(1)], False),
    (list[data.Int], [data.Int.uncheck_new(1.0)], True),  # Runtime incompatible object
    (list[data.Int], [], False),
    (dict[str, int], {"foo": 10}, False),
    # (dict[str, int], {"foo": 10.0}, True),
    (dict[str, data.Int], {"foo": data.Int(1)}, False),
    (dict[str, data.Int], {"foo": data.Int.uncheck_new(1.0)}, True),  # Runtime incompatible object
    (set[int], {10}, False),
    (set[int], {10.0}, True),
    (set[int], [10], True),
    (tuple[int], (10,), False),
    (tuple[int], (10.0,), True),
    (tuple[int, str], (10, "foo"), False),
    (tuple[int, str], (10, 10.0), True),
    (tuple[data.Int, data.Str], (data.Int(1), data.Str("2")), False),
    (tuple[data.Int, data.Str], (data.Int(1), data.Int(2)), True),
    (tuple, (10, 10.0), False),
    (tuple[int, ...], (1, 2), False),
    (data.E, data.E.S, False),
    (data.E, data.IE.V0, True),
    (Union[int, str], 10, False),
    (Union[int, str], "foo", False),
    (Union[int, str], 10.0, True),
    (Union[int, data.Int], data.Int(10), False),
    (datetime.date, datetime.date.today(), False),
    (pathlib.Path, pathlib.Path(), False),
    (pathlib.Path, "foo", True),
]


# Those test cases have wrong runtime values against declared types.
# This is not yet testable until beartype implements O(n) type checking
# https://beartype.readthedocs.io/en/latest/api_decor/#beartype.BeartypeStrategy
default_unstable_test_cases: list[tuple[Any, Any, bool]] = [
    (list[int], [1, 1.0], True),
    (list[data.Int], [data.Int(1), data.Float(10.0)], True),
    (dict[str, int], {"foo": 10, 100: "bar"}, False),
    (tuple[int, ...], (1, 2.0), True),
]


@pytest.mark.parametrize("T,data,exc", test_cases)
def test_type_check_strict(T: Any, data: Any, exc: bool) -> None:
    @serde.serde
    class C:
        a: T  # pyright: ignore[reportInvalidTypeForm]

    if exc:
        with pytest.raises((serde.SerdeError, BeartypeCallHintViolation)):
            d = serde.to_dict(C(data))
            serde.from_dict(C, d)
    else:
        d = serde.to_dict(C(data))
        serde.from_dict(C, d)


def test_type_check_disabled_for_dataclass_without_serde() -> None:
    @dataclass
    class Foo:
        value: int

    f = Foo("100")  # type: ignore
    data = serde.json.to_json(f)
    assert f == serde.json.from_json(Foo, data)

    f = Foo("100")  # type: ignore


def test_uncoercible() -> None:
    @serde.serde(type_check=serde.coerce)
    class Foo:
        i: int

    with pytest.raises(serde.SerdeError):
        serde.to_dict(Foo("foo"))  # type: ignore

    with pytest.raises(serde.SerdeError):
        serde.from_dict(Foo, {"i": "foo"})


def test_coerce() -> None:
    @serde.serde(type_check=serde.coerce)
    class Foo:
        i: int
        s: str
        f: float
        b: bool

    d = {"i": "10", "s": 100, "f": 1000, "b": "True"}
    p = serde.from_dict(Foo, d)
    assert p.i == 10
    assert p.s == "100"
    assert p.f == 1000.0
    assert p.b

    p = Foo("10", 100, 1000, "True")  # type: ignore
    d = serde.to_dict(p)
    assert d["i"] == 10
    assert d["s"] == "100"
    assert d["f"] == 1000.0
    assert d["b"]

    # Couldn't coerce
    with pytest.raises(serde.SerdeError):
        d = {"i": "foo", "s": 100, "f": "bar", "b": "True"}
        p = serde.from_dict(Foo, d)

    @serde.serde(type_check=serde.coerce)
    class Int:
        i: int

    @serde.serde(type_check=serde.coerce)
    class Str:
        s: str

    @serde.serde(type_check=serde.coerce)
    class Float:
        f: float

    @serde.serde(type_check=serde.coerce)
    class Bool:
        b: bool

    @serde.serde(type_check=serde.coerce)
    class Nested:
        i: Int
        s: Str
        f: Float
        b: Bool

    # Nested structure
    p2 = Nested(Int("10"), Str(100), Float(1000), Bool("True"))  # type: ignore
    d2: dict[str, dict[str, Any]] = serde.to_dict(p2)
    assert d2["i"]["i"] == 10
    assert d2["s"]["s"] == "100"
    assert d2["f"]["f"] == 1000.0
    assert d2["b"]["b"]

    d3 = {"i": {"i": "10"}, "s": {"s": 100}, "f": {"f": 1000}, "b": {"b": "True"}}
    p3 = serde.from_dict(Nested, d3)
    assert p3.i.i == 10
    assert p3.s.s == "100"
    assert p3.f.f == 1000.0
    assert p3.b.b

    @serde.serde(type_check=serde.coerce)
    class InnerTuple:
        # Note: `foo` needs to be longer than 1 char, to properly test
        # quote escaping
        foo: tuple[float, float]

    f = InnerTuple(foo=(1, 2))
    assert f.foo == (1.0, 2.0)



from typing import Optional, Union, List

import pytest

import serde as serde_pkg
from serde import serde
from serde.json import deserialize_json_numbers, from_json, to_json


def test_json_basics() -> None:
    @serde
    class Foo:
        v: Optional[int]

    f = Foo(10)
    assert '{"v":10}' == to_json(f)
    assert f == from_json(Foo, '{"v":10}')

    @serde
    class Bar:
        v: set[int]

    b = Bar({1, 2, 3})
    assert '{"v":[1,2,3]}' == to_json(b)
    assert b == from_json(Bar, '{"v":[1,2,3]}')


def test_skip_none() -> None:
    @serde
    class Foo:
        a: int
        b: Optional[int]

    f = Foo(10, 100)
    assert to_json(f, skip_none=True) == '{"a":10,"b":100}'

    f = Foo(10, None)
    assert to_json(f, skip_none=True) == '{"a":10}'


def test_skip_none_nested() -> None:
    """Test that skip_none propagates to nested objects."""

    @serde
    class Inner:
        a: int
        b: Optional[int]

    @serde
    class Outer:
        i: Inner
        j: int
        k: Optional[int]

    # Test nested dataclass with skip_none
    o = Outer(i=Inner(a=1, b=None), j=3, k=None)
    result = to_json(o, skip_none=True)
    expected = '{"i":{"a":1},"j":3}'
    assert result == expected

    # Test union with nested dataclass and skip_none
    @serde
    class WithUnion:
        u: Union[Inner, str]
        value: Optional[str]

    u = WithUnion(u=Inner(a=5, b=None), value=None)
    result = to_json(u, skip_none=True)
    expected = '{"u":{"Inner":{"a":5}}}'
    assert result == expected

    # Test list with nested objects and skip_none
    @serde
    class WithList:
        items: List[Inner]
        name: Optional[str]

    w = WithList(items=[Inner(a=1, b=None), Inner(a=2, b=3)], name=None)
    result = to_json(w, skip_none=True)
    expected = '{"items":[{"a":1},{"a":2,"b":3}]}'
    assert result == expected


def test_deserialize_json_numbers() -> None:
    @serde
    class Foo:
        value: float = serde_pkg.field(deserializer=deserialize_json_numbers)

    assert from_json(Foo, '{"value": 0}').value == 0.0
    assert from_json(Foo, '{"value": 0.5}').value == 0.5

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Foo, '{"value": true}')


def test_coerce_numbers_json() -> None:
    @serde
    class Foo:
        value: float

    foo = from_json(Foo, '{"value": 1}')
    assert foo.value == 1.0

    @serde
    class Bar:
        values: List[float]

    bar = from_json(Bar, '{"values":[1,2]}')
    assert bar.values == [1.0, 2.0]

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Bar, '{"values":[1,"2"]}')

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Foo, '{"value": 1}', coerce_numbers=False)

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Foo, '{"value": true}')


def test_json_numbers_with_union() -> None:
    @serde
    class Foo:
        value: Union[float, int]

    assert from_json(Foo, '{"value": 1}').value == 1.0

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Foo, '{"value": "1"}')

    @serde
    class Bar:
        value: Union[int, float]

    assert from_json(Bar, '{"value": 1}').value == 1

    with pytest.raises(serde_pkg.SerdeError):
        from_json(Bar, '{"value": "1"}')



import jedi


def test_jedi() -> None:
    source = """
from serde import serde

@serde
class Foo:
    a: int
    b: float
    c: str
    baz: bool

foo = Foo(10, 100.0, "foo", True)
"""
    source_completion = source + "\n" + "foo."
    jedi_script = jedi.Script(source_completion, path="foo.py")
    completions = jedi_script.complete(9, len("foo."))
    completions = [comp.name for comp in completions]
    assert "a" in completions
    assert "b" in completions
    assert "c" in completions



import enum
from dataclasses import dataclass
from collections.abc import Callable
from typing import Any, Optional


class Size(enum.Enum):
    Small = "Small"
    Medium = "Medium"
    Large = "Large"


@dataclass
class Runner:
    name: str
    data: Any
    se: Optional[Callable[..., Any]]
    de: Optional[Callable[..., Any]]
    astuple: Optional[Callable[..., Any]]
    asdict: Optional[Callable[..., Any]]



"""
skip.py

Example usage of skip and skip_if attributes.

Usage:
    $ uv sync
    $ uv run python skip.py
"""

from serde import field, serde
from serde.json import to_json


@serde
class Resource:
    name: str
    hash: str
    metadata: dict[str, str] = field(default_factory=dict, skip=True)


@serde
class World:
    player: str
    enemies: list[str] = field(default_factory=list, skip_if_false=True)
    buddy: str = field(default="", skip_if=lambda v: v == "Pikachu")
    town: str = field(default="Masara Town", skip_if_default=True)


def main() -> None:
    resources = [
        Resource("Stack Overflow", "b6469c3f31653d281bbbfa6f94d60fea130abe38"),
        Resource(
            "GitHub",
            "5cb7a0c47e53854cd00e1a968de5abce1c124601",
            metadata={"headquarters": "San Francisco"},
        ),
    ]
    print(to_json(resources))

    # "buddy" and "town" field will be omitted
    world = World("satoshi", ["Rattata", "Pidgey"], "Pikachu")
    print(to_json(world))

    # "enemies" field will be omitted
    world = World("green", [], "Charmander", "Black City")
    print(to_json(world))


if __name__ == "__main__":
    main()



# pyright: reportInvalidTypeForm=false, reportGeneralTypeIssues=false

import dataclasses
import enum
import logging
import uuid
import datetime
from beartype.roar import BeartypeCallHintViolation
from typing import (
    ClassVar,
    Optional,
    Any,
)
from collections import defaultdict

import pytest

import serde
import serde.toml

from . import data
from .common import (
    all_formats,
    format_dict,
    format_json,
    format_msgpack,
    format_pickle,
    format_toml,
    format_tuple,
    format_yaml,
    opt_case,
    opt_case_ids,
    type_ids,
    types,
)

log = logging.getLogger("test")

serde.init(True)


@pytest.mark.parametrize("t,T,f", types, ids=type_ids())
@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_simple(se: Any, de: Any, opt: Any, t: Any, T: Any, f: Any) -> None:
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    if f(se, de, opt):
        return

    @serde.serde(**opt)
    class C:
        i: int
        t: T

    c = C(10, t)
    assert c == de(C, se(c))

    # @serde.serde(**opt)
    # class Nested:
    #    t: T

    # @serde.serde(**opt)
    # class C:
    #    n: Nested

    # c = C(Nested(t))
    # assert c == de(C, se(c))

    # if se is not serde.toml.to_toml:
    #    assert t == de(T, se(t))


@pytest.mark.parametrize("t,T,f", types, ids=type_ids())
@pytest.mark.parametrize("named", (True, False))
@pytest.mark.parametrize("reuse", (True, False))
def test_from_to_obj(t: Any, T: Any, f: Any, named: bool, reuse: bool) -> None:
    obj = serde.se.to_obj(t, named, reuse, False)
    assert t == serde.de.from_obj(T, obj, named, reuse)


@pytest.mark.parametrize("t,T,filter", types, ids=type_ids())
@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", (format_dict + format_tuple))
def test_simple_with_reuse_instances(
    se: Any, de: Any, opt: Any, t: Any, T: Any, filter: Any
) -> None:
    log.info(
        f"Running test with se={se.__name__} de={de.__name__} opts={opt} while reusing instances"
    )

    @serde.serde(**opt)
    class C:
        i: int
        t: T

    c = C(10, t)
    assert c == de(C, se(c, reuse_instances=True), reuse_instances=True)

    @serde.serde(**opt)
    class Nested:
        t: T

    @serde.serde(**opt)
    class C2:
        n: Nested

    c2 = C2(Nested(t))
    assert c2 == de(C2, se(c2, reuse_instances=True), reuse_instances=True)


def test_non_dataclass_reuse_instances() -> None:
    dt = datetime.datetime.fromisoformat("2020-01-01 00:00:00+00:00")
    assert "2020-01-01T00:00:00+00:00" == serde.to_dict(dt, reuse_instances=False)  # type: ignore
    assert dt == serde.to_dict(dt, reuse_instances=True)
    assert dt is serde.to_dict(dt, reuse_instances=True)
    assert "2020-01-01T00:00:00+00:00" == serde.to_tuple(dt, reuse_instances=False)  # type: ignore
    assert dt == serde.to_dict(dt, reuse_instances=True)
    assert dt is serde.to_dict(dt, reuse_instances=True)


def test_non_dataclass() -> None:
    @serde.serde
    class Foo:
        i: int

    @serde.serialize
    class Bar:
        i: int

    @serde.deserialize
    class Baz:
        i: int


# test_string_forward_reference_works currently only works with global visible classes
# and can not be mixed with PEP 563 "from __future__ import annotations"
@dataclasses.dataclass
class ForwardReferenceFoo:
    bar: "ForwardReferenceBar"


@serde.serde
class ForwardReferenceBar:
    i: int


# assert type is str
assert "ForwardReferenceBar" == dataclasses.fields(ForwardReferenceFoo)[0].type

# setup pyserde for Foo after Bar becomes visible to global scope
serde.serde(ForwardReferenceFoo)

# now the type really is of type Bar
assert ForwardReferenceBar == dataclasses.fields(ForwardReferenceFoo)[0].type
assert ForwardReferenceBar == next(serde.compat.dataclass_fields(ForwardReferenceFoo)).type


# verify usage works
def test_string_forward_reference_works() -> None:
    h = ForwardReferenceFoo(bar=ForwardReferenceBar(i=10))
    h_dict = {"bar": {"i": 10}}

    assert serde.to_dict(h) == h_dict
    assert serde.from_dict(ForwardReferenceFoo, h_dict) == h


# trying to use string forward reference normally will throw
def test_unresolved_forward_reference_throws() -> None:
    with pytest.raises(serde.SerdeError) as e:

        @serde.serde
        class UnresolvedForwardFoo:
            bar: "UnresolvedForwardBar"

        @serde.serde
        class UnresolvedForwardBar:
            i: int

    assert "Failed to resolve type hints for UnresolvedForwardFoo" in str(e)


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_list(se: Any, de: Any, opt: Any) -> None:
    @serde.serde(**opt)
    class Variant:
        d: list[Any]

    # List can contain different types (except Toml).
    if se is not serde.toml.to_toml:
        p = Variant([10, "foo", 10.0, True])
        assert p == de(Variant, se(p))


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_dict_with_non_str_keys(se: Any, de: Any, opt: Any) -> None:
    @serde.serde(**opt)
    class Foo:
        i: dict[int, int]
        s: dict[str, str]
        f: dict[float, float]
        b: dict[bool, bool]

    # JSON, Msgpack, Toml don't allow non string key.
    if se not in (serde.json.to_json, serde.msgpack.to_msgpack, serde.toml.to_toml):
        p = Foo({10: 10}, {"foo": "bar"}, {100.0: 100.0}, {True: False})
        assert p == de(Foo, se(p))


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_enum(se: Any, de: Any, opt: Any) -> None:
    from serde.compat import is_enum

    from .data import IE, E

    class Inner(enum.IntEnum):
        V0 = enum.auto()
        V1 = enum.auto()
        V2 = enum.auto()

    class NestedEnum(enum.Enum):
        V = Inner.V0

    @serde.serde(**opt)
    class Foo:
        a: E
        b: IE
        c: NestedEnum
        d: E = E.S
        e: IE = IE.V1
        f: NestedEnum = NestedEnum.V

    f = Foo(E.S, IE.V0, NestedEnum.V)
    ff = de(Foo, se(f))
    assert f == ff
    assert is_enum(ff.a) and isinstance(ff.a, E)
    assert is_enum(ff.b) and isinstance(ff.b, IE)
    assert is_enum(ff.c) and isinstance(ff.c, NestedEnum)
    assert is_enum(ff.d) and isinstance(ff.d, E)
    assert is_enum(ff.e) and isinstance(ff.e, IE)
    assert is_enum(ff.f) and isinstance(ff.f, NestedEnum)

    f = Foo(E("foo"), IE(2), NestedEnum(Inner.V0), E(True), IE(10), NestedEnum(Inner.V0))
    try:
        data = se(f)
    except Exception:
        pass
    else:
        ff = de(Foo, data)
        assert is_enum(ff.a) and isinstance(ff.a, E) and ff.a == E.S
        assert is_enum(ff.b) and isinstance(ff.b, IE) and ff.b == IE.V1
        assert is_enum(ff.c) and isinstance(ff.c, NestedEnum) and ff.c == NestedEnum.V
        assert is_enum(ff.d) and isinstance(ff.d, E) and ff.d == E.B
        assert is_enum(ff.e) and isinstance(ff.e, IE) and ff.e == IE.V2
        assert is_enum(ff.f) and isinstance(ff.f, NestedEnum) and ff.f == NestedEnum.V


# TODO
# @pytest.mark.parametrize("se,de", all_formats)
# def test_enum_imported(se, de):
#     from .data import EnumInClass
#
#     c = EnumInClass()
#     cc = de(EnumInClass, se(c))
#     assert c == cc


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_tuple(se: Any, de: Any, opt: Any) -> None:
    @serde.serde(**opt)
    class Homogeneous:
        i: tuple[int, int]
        s: tuple[str, str]
        f: tuple[float, float]
        b: tuple[bool, bool]

    def uncheck_new(i: list[Any], s: list[Any], f: list[Any], b: list[Any]) -> Homogeneous:
        """
        Bypass runtime type checker by mutating inner value.
        """
        obj = Homogeneous((0, 0), ("", ""), (0.0, 0.0), (True, True))
        obj.i = i  # type: ignore
        obj.s = s  # type: ignore
        obj.f = f  # type: ignore
        obj.b = b  # type: ignore
        return obj

    a = Homogeneous((10, 20), ("a", "b"), (10.0, 20.0), (True, False))
    assert a == de(Homogeneous, se(a))

    # List will be type mismatch if type_check=True.
    tuple_but_actually_list = uncheck_new([10, 20], ["a", "b"], [10.0, 20.0], [True, False])
    assert tuple_but_actually_list != de(Homogeneous, se(tuple_but_actually_list))

    @serde.serde(**opt)
    class Variant:
        t: tuple[int, str, float, bool]

    # Toml doesn't support variant type of array.
    if se is not serde.toml.to_toml:
        b = Variant((10, "a", 10.0, True))
        assert b == de(Variant, se(b))

    @serde.serde(**opt)
    class Baretuple:
        t: tuple  # type: ignore

    c = Baretuple((10, 20))
    assert c == de(Baretuple, se(c))

    c = Baretuple(())
    assert c == de(Baretuple, se(c))

    @serde.serde(**opt)
    class Nested:
        i: tuple[data.Int, data.Int]
        s: tuple[data.Str, data.Str]
        f: tuple[data.Float, data.Float]
        b: tuple[data.Bool, data.Bool]

    d = Nested(
        (data.Int(10), data.Int(20)),
        (data.Str("a"), data.Str("b")),
        (data.Float(10.0), data.Float(20.0)),
        (data.Bool(True), data.Bool(False)),
    )
    assert d == de(Nested, se(d))

    @serde.serde(**opt)
    class Variabletuple:
        t: tuple[int, int, int]
        i: tuple[data.Inner, data.Inner]

    e = Variabletuple((1, 2, 3), (data.Inner(0), data.Inner(1)))
    assert e == de(Variabletuple, se(e))

    with pytest.raises((serde.SerdeError, BeartypeCallHintViolation)):
        e = Variabletuple((), ())  # type: ignore[arg-type]
        assert e == de(Variabletuple, se(e))

    with pytest.raises((serde.SerdeError, SyntaxError)):

        @serde.serde(**opt)
        class Emptytuple:
            t: tuple[()]


@pytest.mark.parametrize("se,de", all_formats)
def test_single_element_tuples(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        a: tuple[int]
        b: tuple[uuid.UUID]

    foo = Foo(a=(1,), b=(uuid.UUID("855f07da-c3cd-46f2-9557-b8dbeb99ff42"),))
    assert serde.to_dict(foo) == {"a": foo.a, "b": foo.b}

    assert foo == de(Foo, se(foo))


@pytest.mark.parametrize("se,de", all_formats)
def test_dataclass_default_factory(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        foo: str
        items: dict[str, int] = serde.field(default_factory=dict)

    f = Foo("bar")
    assert f == de(Foo, se(f))

    assert {"foo": "bar", "items": {}} == serde.to_dict(f)
    assert f == serde.from_dict(Foo, {"foo": "bar"})


@pytest.mark.parametrize("se,de", all_formats)
def test_default(se: Any, de: Any) -> None:
    from serde import from_dict, from_tuple

    from .data import OptDefault, PriDefault

    p = PriDefault()
    assert p == de(PriDefault, se(p))

    p = PriDefault()
    assert p == from_dict(PriDefault, {})
    assert p == from_dict(PriDefault, {"i": 10})
    assert p == from_dict(PriDefault, {"i": 10, "s": "foo"})
    assert p == from_dict(PriDefault, {"i": 10, "s": "foo", "f": 100.0})
    assert p == from_dict(PriDefault, {"i": 10, "s": "foo", "f": 100.0, "b": True})
    assert p == from_tuple(PriDefault, (10, "foo", 100.0, True))

    if se is not serde.toml.to_toml:  # Toml doesn't support None.
        o = OptDefault()
        assert o == de(OptDefault, se(o))
        o = OptDefault()
        assert o == from_dict(OptDefault, {})
        assert o == from_dict(OptDefault, {"n": None})
        assert o == from_dict(OptDefault, {"n": None, "i": 10})
        assert o == from_tuple(OptDefault, (None, 10))

    if se is not serde.toml.to_toml:
        o = OptDefault(n=None, i=None)
        assert o == from_dict(OptDefault, {"n": None, "i": None})
        assert o == from_tuple(OptDefault, (None, None))

    assert 10 == dataclasses.fields(PriDefault)[0].default
    assert "foo" == dataclasses.fields(PriDefault)[1].default
    assert 100.0 == dataclasses.fields(PriDefault)[2].default
    assert True is dataclasses.fields(PriDefault)[3].default


@pytest.mark.parametrize(
    "se,de",
    (format_dict + format_tuple + format_json + format_msgpack + format_yaml + format_pickle),
)
def test_list_pri(se: Any, de: Any) -> None:
    p = [data.PRI, data.PRI]
    assert p == de(data.ListPri, se(p))

    p = []
    assert p == de(data.ListPri, se(p))


@pytest.mark.parametrize(
    "se,de",
    (format_dict + format_tuple + format_json + format_msgpack + format_yaml + format_pickle),
)
def test_dict_pri(se: Any, de: Any) -> None:
    p = {"1": data.PRI, "2": data.PRI}
    assert p == de(data.DictPri, se(p))

    p = {}
    assert p == de(data.DictPri, se(p))


def test_json() -> None:
    p = data.Pri(10, "foo", 100.0, True)
    s = '{"i":10,"s":"foo","f":100.0,"b":true}'
    assert s == serde.json.to_json(p)
    assert '"😊"' == serde.json.to_json("😊")
    assert "10" == serde.json.to_json(10)
    assert "[10,20,30]" == serde.json.to_json([10, 20, 30])
    assert '{"foo":10,"fuga":10}' == serde.json.to_json({"foo": 10, "fuga": 10})


def test_msgpack() -> None:
    p = data.Pri(10, "foo", 100.0, True)
    d = b"\x84\xa1i\n\xa1s\xa3foo\xa1f\xcb@Y\x00\x00\x00\x00\x00\x00\xa1b\xc3"
    assert d == serde.msgpack.to_msgpack(p)
    assert p == serde.msgpack.from_msgpack(data.Pri, d)


def test_msgpack_unnamed() -> None:
    p = data.Pri(10, "foo", 100.0, True)
    d = b"\x94\n\xa3foo\xcb@Y\x00\x00\x00\x00\x00\x00\xc3"
    assert d == serde.msgpack.to_msgpack(p, named=False)
    assert p == serde.msgpack.from_msgpack(data.Pri, d, named=False)


@pytest.mark.parametrize("se,de", all_formats)
def test_rename(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        class_name: str = serde.field(rename="class")

    f = Foo(class_name="foo")
    assert f == de(Foo, se(f))


@pytest.mark.parametrize("se,de", format_msgpack)
def test_rename_msgpack(se: Any, de: Any) -> None:
    @serde.serde(rename_all="camelcase")
    class Foo:
        class_name: str

    f = Foo(class_name="foo")
    assert f == de(Foo, se(f, named=True), named=True)
    assert f == de(Foo, se(f, named=False), named=False)


@pytest.mark.parametrize(
    "se,de", (format_dict + format_json + format_yaml + format_toml + format_pickle)
)
def test_rename_formats(se: Any, de: Any) -> None:
    @serde.serde(rename_all="camelcase")
    class Foo:
        class_name: str

    f = Foo(class_name="foo")
    assert f == de(Foo, se(f))


@pytest.mark.parametrize(
    "se,de", (format_dict + format_json + format_yaml + format_toml + format_pickle)
)
def test_alias(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        a: str = serde.field(alias=["b", "c", "d"])  # type: ignore[literal-required]

    f = Foo(a="foo")
    assert f == de(Foo, se(f))


def test_conflicting_alias() -> None:
    @serde.serde
    class Foo:
        a: int = serde.field(alias=["b", "c", "d"])  # type: ignore[literal-required]
        b: int
        c: int
        d: int

    f = Foo(a=1, b=2, c=3, d=4)
    assert '{"a":1,"b":2,"c":3,"d":4}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"a":1,"b":2,"c":3,"d":4}')
    assert ff.a == 1
    assert ff.b == 2
    assert ff.c == 3
    assert ff.d == 4

    ff = serde.json.from_json(Foo, '{"b":2,"c":3,"d":4}')
    assert ff.a == 2
    assert ff.b == 2
    assert ff.c == 3
    assert ff.d == 4


def test_rename_and_alias() -> None:
    @serde.serde
    class Foo:
        a: int = serde.field(rename="z", alias=["b", "c", "d"])  # type: ignore[literal-required]

    f = Foo(a=1)
    assert '{"z":1}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"b":10}')
    assert ff.a == 10


def test_rename_all_and_alias() -> None:
    @serde.serde(rename_all="pascalcase")
    class Foo:
        a_field: int = serde.field(alias=["b_field"])  # type: ignore[literal-required]

    f = Foo(1)
    assert '{"AField":1}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"b_field":1}')  # alias is not renamed
    assert f == ff


def test_default_and_alias() -> None:
    @serde.serde
    class Foo:
        a: int = serde.field(default=2, alias=["b", "c", "d"])  # type: ignore[literal-required]

    f = Foo(a=1)
    assert '{"a":1}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"b":10}')
    assert ff.a == 10
    ff = serde.json.from_json(Foo, '{"e":10}')
    assert ff.a == 2


def test_optional_and_alias() -> None:
    @serde.serde
    class Foo:
        a: Optional[int] = serde.field(alias=["b"])  # type: ignore[literal-required]

    assert Foo(1) == serde.json.from_json(Foo, '{"b":1}')
    assert Foo(None) == serde.json.from_json(Foo, '{"c":1}')


def test_default_and_rename() -> None:
    @serde.serde
    class Foo:
        a: int = serde.field(default=2, rename="z")

    f = Foo(a=1)
    assert '{"z":1}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"z":10}')
    assert ff.a == 10
    fff = serde.json.from_json(Foo, '{"a":10}')
    assert fff.a == 2


def test_default_rename_and_alias() -> None:
    @serde.serde
    class Foo:
        a: int = serde.field(default=2, rename="z", alias=["b", "c", "d"])  # type: ignore[literal-required]

    f = Foo(a=1)
    assert '{"z":1}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"b":10}')
    assert ff.a == 10
    ff = serde.json.from_json(Foo, '{"e":10}')
    assert ff.a == 2


@pytest.mark.parametrize(
    "se,de",
    (format_dict + format_json + format_msgpack + format_yaml + format_toml + format_pickle),
)
def test_skip_if(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        comments: Optional[list[str]] = serde.field(
            default_factory=list, skip_if=lambda v: len(v) == 0
        )
        attrs: Optional[dict[str, str]] = serde.field(
            default_factory=dict, skip_if=lambda v: v is None or len(v) == 0
        )

    f = Foo(["foo"], {"bar": "baz"})
    assert f == de(Foo, se(f))

    f = Foo([])
    ff = de(Foo, se(f))
    assert ff.comments == []
    assert ff.attrs == {}


@pytest.mark.parametrize("se,de", all_formats)
def test_skip_if_false(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        comments: Optional[list[str]] = serde.field(default_factory=list, skip_if_false=True)

    f = Foo(["foo"])
    assert f == de(Foo, se(f))


@pytest.mark.parametrize(
    "se,de",
    (format_dict + format_json + format_msgpack + format_yaml + format_toml + format_pickle),
)
def test_skip_if_overrides_skip_if_false(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        comments: Optional[list[str]] = serde.field(
            default_factory=list, skip_if_false=True, skip_if=lambda v: len(v) == 1
        )

    f = Foo(["foo"])
    ff = de(Foo, se(f))
    assert ff.comments == []


@pytest.mark.parametrize("se,de", all_formats)
def test_skip_if_default(se: Any, de: Any) -> None:
    @serde.serde
    class Foo:
        a: str = serde.field(default="foo", skip_if_default=True)

    f = Foo()
    assert f == de(Foo, se(f))

    assert serde.to_dict(Foo()) == {}
    assert serde.to_dict(Foo("bar")) == {"a": "bar"}
    assert serde.from_dict(Foo, {}) == Foo()
    assert serde.from_dict(Foo, {"a": "bar"}) == Foo("bar")


@pytest.mark.parametrize("se,de", format_msgpack)
def test_inheritance(se: Any, de: Any) -> None:
    @serde.serde
    class Base:
        a: int
        b: str

    @serde.serde
    class Derived(Base):
        c: float

    base = Base(10, "foo")
    assert base == de(Base, se(base))

    derived = Derived(10, "foo", 100.0)
    assert derived == de(Derived, se(derived))


@pytest.mark.parametrize("se,de", format_msgpack)
def test_duplicate_decorators(se: Any, de: Any) -> None:
    from dataclasses import dataclass

    @serde.serde
    @serde.serde
    @dataclass
    @dataclass
    class Foo:
        a: int
        b: str

    foo = Foo(10, "foo")
    assert foo == de(Foo, se(foo))


@pytest.mark.parametrize("se,de", format_msgpack)
def test_ext(se: Any, de: Any) -> None:
    @serde.serde
    class Base:
        i: int
        s: str

    @serde.serde
    class DerivedA(Base):
        j: int

    @serde.serde
    class DerivedB(Base):
        k: float

    a = DerivedA(i=7, s="A", j=13)
    aa = de(Base, se(a))
    assert aa != a

    EXT_TYPE_DICT = {0: DerivedA, 1: DerivedB}
    # reverse the external type dict for faster serialization
    EXT_TYPE_DICT_REVERSED = {v: k for k, v in EXT_TYPE_DICT.items()}

    a = DerivedA(i=7, s="A", j=13)
    aa = de(None, se(a, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict=EXT_TYPE_DICT)
    assert aa == a

    b = DerivedB(i=3, s="B", k=11.0)
    bb = de(None, se(b, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict=EXT_TYPE_DICT)
    assert b == bb

    with pytest.raises(serde.SerdeError) as se_ex:
        se(a, ext_dict={})
    assert str(se_ex.value) == "Could not find type code for DerivedA in ext_dict"

    with pytest.raises(serde.SerdeError) as de_ex:
        de(None, se(a, ext_dict=EXT_TYPE_DICT_REVERSED), ext_dict={})
    assert str(de_ex.value) == "Could not find type for code 0 in ext_dict"


def test_exception_on_not_supported_types() -> None:
    class UnsupportedClass:
        def __init__(self) -> None:
            pass

    @serde.serde
    class Foo:
        b: UnsupportedClass

    with pytest.raises(serde.SerdeError) as se_ex:
        serde.to_dict(Foo(UnsupportedClass()))
    assert str(se_ex.value).startswith("Unsupported type: UnsupportedClass")

    with pytest.raises(serde.SerdeError) as de_ex:
        serde.from_dict(Foo, {"b": UnsupportedClass()})
    assert str(de_ex.value).startswith("Unsupported type: UnsupportedClass")


def test_dataclass_inheritance() -> None:
    @serde.serde
    class Base:
        i: int
        s: str

    @serde.serde
    class DerivedA(Base):
        j: int

    @serde.serde
    class DerivedB(Base):
        k: float

    # each class should have own scope
    # ensure the generated code of DerivedB does not overwrite the earlier generated code
    # from DerivedA
    assert getattr(Base, serde.core.SERDE_SCOPE) is not getattr(DerivedA, serde.core.SERDE_SCOPE)
    assert getattr(DerivedA, serde.core.SERDE_SCOPE) is not getattr(
        DerivedB, serde.core.SERDE_SCOPE
    )

    base = Base(i=0, s="foo")
    assert base == serde.from_dict(Base, serde.to_dict(base))

    a = DerivedA(i=0, s="foo", j=42)
    assert a == serde.from_dict(DerivedA, serde.to_dict(a))

    b = DerivedB(i=0, s="foo", k=42.0)
    assert b == serde.from_dict(DerivedB, serde.to_dict(b))


def make_serde(class_name: str, se: bool, fields: Any, *args: Any, **kwargs: Any) -> Any:
    if se:
        return serde.de.deserialize(
            serde.se._make_serialize(class_name, fields, *args, **kwargs), **kwargs
        )
    else:
        return serde.se.serialize(
            serde.de._make_deserialize(class_name, fields, *args, **kwargs), **kwargs
        )


@pytest.mark.parametrize("se", (True, False))
def test_make_serialize_deserialize(se: Any) -> None:
    fields = [("i", int, dataclasses.field())]
    Foo = make_serde("Foo", se, fields)

    f = Foo(10)
    assert serde.to_dict(f) == {"i": 10}
    assert str(f) == "Foo(i=10)"
    assert serde.from_dict(Foo, {"i": 10}) == f

    # Test class attribute
    fields = [("int_field", int, dataclasses.field())]
    Foo = make_serde("Foo", se, fields, rename_all="pascalcase")
    f = Foo(10)
    assert serde.to_dict(f) == {"IntField": 10}

    # Test field attribute
    fields = [
        ("i", int, dataclasses.field(metadata={"serde_skip": True})),
        ("j", float, dataclasses.field()),  # type: ignore[list-item]
    ]
    Foo = make_serde("Foo", se, fields)
    f = Foo(10, 100.0)
    assert serde.to_dict(f) == {"j": 100.0}

    # Test class/field attributes at the same time
    fields = [
        ("int_field", int, dataclasses.field(metadata={"serde_rename": "renamed_field"})),
        ("float_field", float, dataclasses.field()),  # type: ignore[list-item]
    ]
    Foo = make_serde("Foo", se, fields, rename_all="pascalcase")
    f = Foo(10, 100.0)
    assert serde.to_dict(f) == {"renamed_field": 10, "FloatField": 100.0}

    # Nested
    fields = [("v", int, dataclasses.field())]
    Bar = make_serde("Bar", se, fields)

    fields = [("bar", Bar, dataclasses.field())]
    Foo = make_serde("Foo", se, fields)
    f = Foo(Bar(10))
    assert serde.to_dict(f) == {"bar": {"v": 10}}
    assert serde.from_dict(Foo, {"bar": {"v": 10}}) == f


def test_exception_to_from_obj() -> None:
    @serde.serde
    class Foo:
        a: int

    class Bar:
        pass

    with pytest.raises(serde.SerdeError):
        serde.from_dict(Foo, {})


def test_user_error() -> None:
    class MyException(Exception):
        pass

    @serde.serde
    class Foo:
        v: int

        def __post_init__(self) -> None:
            if self.v == 10:
                raise MyException("Invalid value")

    with pytest.raises(MyException):
        serde.from_dict(Foo, {"v": 10})

    with pytest.raises(serde.SerdeError):
        serde.from_dict(Foo, {})


def test_frozenset() -> None:
    @serde.serde
    class Foo:
        d: frozenset[int]

    f = Foo(frozenset({1, 2}))
    assert '{"d":[1,2]}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"d":[1,2]}')
    assert f == ff
    assert isinstance(f.d, frozenset)
    assert isinstance(ff.d, frozenset)

    fs = serde.json.from_json(frozenset[int], "[1,2]")
    assert fs == frozenset([1, 2])


def test_defaultdict() -> None:
    from collections import defaultdict

    @serde.serde
    class Foo:
        v: defaultdict[str, list[int]]

    f = Foo(defaultdict(list, {"k": [1, 2]}))
    assert '{"v":{"k":[1,2]}}' == serde.json.to_json(f)
    ff = serde.json.from_json(Foo, '{"v":{"k":[1,2]}}')
    assert f == ff
    assert isinstance(f.v, defaultdict)
    assert isinstance(ff.v, defaultdict)

    dd = serde.json.from_json(defaultdict[str, list[int]], '{"k":[1,2]}')
    assert isinstance(dd, defaultdict)
    assert dd == defaultdict(list, {"k": [1, 2]})


def test_defaultdict_invalid_value_type() -> None:
    with pytest.raises(serde.SerdeError):
        serde.json.from_json(defaultdict[str, ...], '{"k":[1,2]}')  # type: ignore[misc]

    with pytest.raises(serde.SerdeError):
        serde.json.from_json(defaultdict, '{"k":[1,2]}')


def test_class_var() -> None:
    @serde.serde
    class Disabled:
        v: ClassVar[int] = 10

    a = Disabled()
    assert {} == serde.to_dict(a)
    assert a == serde.from_dict(Disabled, {})

    @serde.serde(serialize_class_var=True)
    class Enabled:
        v: ClassVar[int] = 10

    b = Enabled()
    assert {"v": 10} == serde.to_dict(b)
    assert b == serde.from_dict(Enabled, {})

    @serde.serde
    class Foo:
        v: int

    @serde.serde(serialize_class_var=True)
    class Dataclass:
        v: ClassVar[Foo] = Foo(10)

    c = Dataclass()
    assert {"v": {"v": 10}} == serde.to_dict(c)
    assert c == serde.from_dict(Dataclass, {})

    @serde.serde(serialize_class_var=True)
    class ClassList:
        v: ClassVar[list[Foo]] = [Foo(10), Foo(20)]

    d = ClassList()
    assert {"v": [{"v": 10}, {"v": 20}]} == serde.to_dict(d)
    assert d == serde.from_dict(ClassList, {})

    # Mutate class variable
    Enabled.v = 100
    e = Enabled()
    assert {"v": 100} == serde.to_dict(e)
    assert e == serde.from_dict(Enabled, {})

    @serde.serde(serialize_class_var=True)
    class Bar:
        v: ClassVar[int] = 100

    @serde.serde(serialize_class_var=True)
    class Nested:
        v: ClassVar[Bar] = Bar()

    f = Nested()
    assert {"v": {"v": 100}} == serde.to_dict(f)
    assert f == serde.from_dict(Nested, {})


def test_dataclass_without_serde() -> None:
    @serde.serde(rename_all="kebabcase")
    class Foo:
        myLongName: int

    @dataclasses.dataclass
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(myLongName=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"my-long-name": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_dataclass_add_serialize() -> None:
    @serde.serde(rename_all="kebabcase")
    class Foo:
        myLongName: int

    @serde.deserialize
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(myLongName=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"my-long-name": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_dataclass_add_deserialize() -> None:
    @serde.serde(rename_all="kebabcase")
    class Foo:
        myLongName: int

    @serde.serialize
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(myLongName=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"my-long-name": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_nested_dataclass_without_serde() -> None:
    @dataclasses.dataclass
    class Foo:
        v: int

    @serde.serde
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(v=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"v": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_nested_dataclass_add_deserialize() -> None:
    @serde.serialize
    class Foo:
        v: int

    @serde.serde
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(v=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"v": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_nested_dataclass_add_serialize() -> None:
    @serde.deserialize
    class Foo:
        v: int

    @serde.serde
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(v=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"v": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_nested_dataclass_ignore_wrapper_options() -> None:
    @dataclasses.dataclass
    class Foo:
        myLongName: int

    @serde.serde(rename_all="kebabcase")
    class Wrapper:
        foo: Foo

    a = Wrapper(foo=Foo(myLongName=1))
    serialized = serde.to_dict(a)
    assert {"foo": {"myLongName": 1}} == serialized
    assert a == serde.from_dict(Wrapper, serialized)


def test_deserialize_from_incompatible_value() -> None:
    @serde.deserialize
    class Foo:
        v: int

    with pytest.raises(serde.SerdeError):
        assert serde.from_dict(Foo, None)  # type: ignore[call-overload]
    with pytest.raises(serde.SerdeError):
        assert serde.from_dict(Foo, "")  # type: ignore[call-overload]
    with pytest.raises(serde.SerdeError):
        assert serde.from_dict(Foo, 10)  # type: ignore[call-overload]


def test_frozen_dataclass() -> None:
    @serde.serde
    @dataclasses.dataclass(frozen=True)
    class Foo:
        a: int

    @serde.serde
    @dataclasses.dataclass(frozen=True)
    class Bar(Foo):  # type: ignore[misc]
        b: int

    bar = Bar(a=10, b=20)
    assert bar == serde.from_dict(Bar, serde.to_dict(bar))


def test_init_var_with_field_attribute() -> None:
    @serde.serde
    @dataclasses.dataclass
    class Foo:
        a: int = serde.field(skip=True)

    assert serde.to_dict(Foo(10)) == {}


def test_dict_str_any() -> None:
    @serde.serde
    class Foo:
        a: dict[str, Any]

    # Because the dict values are typed as Any, all values will be serialized as strings
    date = datetime.datetime(2024, 12, 3, 16, 55, 20, 220662)
    date_str = date.isoformat()

    uuid_val = uuid.UUID("efb71d76-4337-4b27-a612-5b6f95b01d42")
    uuid_str = str(uuid_val)

    foo = Foo(a={"a": uuid_val, "b": date})
    foo_se = {"a": {"a": uuid_str, "b": date_str}}
    foo_de = Foo(a={"a": uuid_str, "b": date_str})

    assert serde.to_dict(foo) == foo_se
    assert serde.from_dict(Foo, foo_se) == foo_de



# type: ignore

import logging
from dataclasses import fields
from typing import Any

import numpy as np
import numpy.typing as npt
import jaxtyping
import pytest

import serde
from serde.compat import get_origin

from .common import all_formats, format_json, format_msgpack, opt_case, opt_case_ids

log = logging.getLogger("test")

serde.init(True)


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_numpy_simple(se: Any, de: Any, opt: Any) -> None:
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    @serde.serde(**opt)
    class NPFoo:
        i: np.int32
        j: np.int64
        f: np.float32
        g: np.float64
        h: np.bool_
        u: np.ndarray  # type: ignore[type-arg]
        v: npt.NDArray  # type: ignore[type-arg]
        w: npt.NDArray[np.int32]
        x: npt.NDArray[np.int64]
        y: npt.NDArray[np.float32]
        z: npt.NDArray[np.float64]

        def __eq__(self, other: Any) -> bool:
            return bool(
                self.i == other.i
                and self.j == other.j
                and self.f == other.f
                and self.g == other.g
                and self.h == other.h
                and (self.u == other.u).all()
                and (self.v == other.v).all()
                and (self.w == other.w).all()
                and (self.x == other.x).all()
                and (self.y == other.y).all()
                and (self.z == other.z).all()
            )

    npfoo = NPFoo(
        np.int32(1),
        np.int64(2),
        np.float32(3.0),
        np.float64(4.0),
        np.bool_(False),
        np.array([1, 2]),
        np.array([3, 4]),
        np.array([np.int32(i) for i in [1, 2]]),
        np.array([np.int64(i) for i in [3, 4]]),
        np.array([np.float32(i) for i in [5.0, 6.0]]),
        np.array([np.float64(i) for i in [7.0, 8.0]]),
    )

    de_npfoo = de(NPFoo, se(npfoo))

    assert npfoo == de_npfoo

    for field in fields(NPFoo):
        value = getattr(npfoo, field.name)
        de_value = getattr(de_npfoo, field.name)

        assert type(value) is type(de_value)

        typ = field.type
        orig_type = get_origin(typ)
        if orig_type is not None:
            typ = orig_type

        assert type(de_value) is typ
        assert type(de_value).dtype == field.type.dtype

    @serde.serde(**opt)
    class NumpyDate:
        d: np.datetime64

    date_test = NumpyDate(np.datetime64(10, "Y"))

    assert de(NumpyDate, se(date_test)) == date_test

    @serde.serde(**opt)
    class NumpyJaxtyping:
        float16: jaxtyping.Float16[np.ndarray, "2 2"]  # noqa: F722
        float32: jaxtyping.Float32[np.ndarray, "2 2"]  # noqa: F722
        float64: jaxtyping.Float64[np.ndarray, "2 2"]  # noqa: F722
        inexact: jaxtyping.Inexact[np.ndarray, "2 2"]  # noqa: F722
        int_: jaxtyping.Int[np.ndarray, "2 2"]  # noqa: F722
        int8: jaxtyping.Int8[np.ndarray, "2 2"]  # noqa: F722
        int16: jaxtyping.Int16[np.ndarray, "2 2"]  # noqa: F722
        int32: jaxtyping.Int32[np.ndarray, "2 2"]  # noqa: F722
        int64: jaxtyping.Int64[np.ndarray, "2 2"]  # noqa: F722
        integer: jaxtyping.Integer[np.ndarray, "2 2"]  # noqa: F722
        uint: jaxtyping.UInt[np.ndarray, "2 2"]  # noqa: F722
        uint8: jaxtyping.UInt8[np.ndarray, "2 2"]  # noqa: F722
        uint16: jaxtyping.UInt16[np.ndarray, "2 2"]  # noqa: F722
        uint32: jaxtyping.UInt32[np.ndarray, "2 2"]  # noqa: F722
        uint64: jaxtyping.UInt64[np.ndarray, "2 2"]  # noqa: F722

        def __eq__(self, other):
            return (
                (self.float16 == other.float16).all()
                and (self.float32 == other.float32).all()
                and (self.float64 == other.float64).all()
                and (self.inexact == other.inexact).all()
                and (self.int_ == other.int_).all()
                and (self.int8 == other.int8).all()
                and (self.int16 == other.int16).all()
                and (self.int32 == other.int32).all()
                and (self.int64 == other.int64).all()
                and (self.integer == other.integer).all()
                and (self.uint == other.uint).all()
                and (self.uint8 == other.uint8).all()
                and (self.uint16 == other.uint16).all()
                and (self.uint32 == other.uint32).all()
                and (self.uint64 == other.uint64).all()
            )

    jaxtyping_test = NumpyJaxtyping(
        float16=np.array([[5, 6], [7, 8]], dtype=np.float16),
        float32=np.array([[9, 10], [11, 12]], dtype=np.float32),
        float64=np.array([[13, 14], [15, 16]], dtype=np.float64),
        inexact=np.array([[17, 18], [19, 20]], dtype=np.float64),
        int_=np.array([[21, 22], [23, 24]], dtype=np.int_),
        int8=np.array([[25, 26], [27, 28]], dtype=np.int8),
        int16=np.array([[29, 30], [31, 32]], dtype=np.int16),
        int32=np.array([[33, 34], [35, 36]], dtype=np.int32),
        int64=np.array([[37, 38], [39, 40]], dtype=np.int64),
        integer=np.array([[41, 42], [43, 44]], dtype=np.int_),
        uint=np.array([[45, 46], [47, 48]], dtype=np.uint),
        uint8=np.array([[49, 50], [51, 52]], dtype=np.uint8),
        uint16=np.array([[53, 54], [55, 56]], dtype=np.uint16),
        uint32=np.array([[57, 58], [59, 60]], dtype=np.uint32),
        uint64=np.array([[61, 62], [63, 64]], dtype=np.uint64),
    )

    assert de(NumpyJaxtyping, se(jaxtyping_test)) == jaxtyping_test


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", format_json + format_msgpack)
def test_encode_numpy(se, de, opt):
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    de_int = de(int, se(np.int32(1)))
    assert de_int == 1
    assert isinstance(de_int, int)

    de_arr = de(list, se(np.array([1, 2, 3])))
    assert isinstance(de_arr, list)
    assert de_arr == [1, 2, 3]

    @serde.serde(**opt)
    class MisTyped:
        a: int
        b: float
        h: bool
        c: list[int]
        d: list[float]
        e: list[bool]

    def unchecked_new(
        a: np.int32,
        b: np.float32,
        h: np.bool_,
        c: npt.NDArray[np.int32],
        d: npt.NDArray[np.float32],
        e: npt.NDArray[np.bool_],
    ) -> MisTyped:
        obj = MisTyped(
            0,
            0.0,
            True,
            [],
            [],
            [],
        )
        obj.a = a  # type: ignore
        obj.b = b  # type: ignore
        obj.h = h  # type: ignore
        obj.c = c  # type: ignore
        obj.d = d  # type: ignore
        obj.e = e  # type: ignore
        return obj

    expected = MisTyped(1, 3.0, False, [1, 2], [5.0, 6.0], [True, False])

    test1 = unchecked_new(
        np.int32(1),
        np.float32(3.0),
        np.bool_(False),
        np.array([np.int32(i) for i in [1, 2]]),
        np.array([np.float32(i) for i in [5.0, 6.0]]),
        np.array([np.bool_(i) for i in [True, False]]),
    )

    assert de(MisTyped, se(test1)) == expected

    test2 = unchecked_new(
        np.int64(1),
        np.float64(3.0),
        np.bool_(False),
        np.array([np.int64(i) for i in [1, 2]]),
        np.array([np.float64(i) for i in [5.0, 6.0]]),
        np.array([np.bool_(i) for i in [True, False]]),
    )

    assert de(MisTyped, se(test2)) == expected

    test3 = unchecked_new(
        np.int64(1),
        np.float64(3.0),
        np.bool_(False),
        np.array([np.int64(i) for i in [1, 2]]),
        np.array([np.float64(i) for i in [5.0, 6.0]]),
        np.array([np.bool_(i) for i in [True, False]]),
    )

    assert de(MisTyped, se(test3)) == expected

    np_datetime = np.datetime64("2022-04-27")
    assert de(np.datetime64, se(np_datetime)) == np_datetime

    for value in [np.int32(1), np.int64(1), np.bool_(False), np.bool_(True), int(1), False, True]:
        typ = type(value)
        de_value = de(typ, se(value))
        assert de_value == value
        assert type(de_value) is typ

        # test bare numpy array deserialization
        arr = np.array([0, 1, 2], dtype=typ)
        de_arr = de(np.ndarray, se(arr))
        assert (de_arr == arr).all()

        # test arrays with dtype=type(value)
        arr_typ = npt.NDArray[typ]

        de_arr = de(arr_typ, se(arr))
        assert (de_arr == arr).all()
        assert de_arr.dtype == arr.dtype == typ

    class BadClass:
        def __init__(self, x):
            self.x = x

    b = BadClass(1)

    with pytest.raises(TypeError):
        se(b)


@pytest.mark.parametrize("se,de", format_json + format_msgpack)
def test_encode_numpy_with_no_default_encoder(se, de):
    log.info(f"Running test with se={se.__name__} de={de.__name__} with no default encoder")

    with pytest.raises(TypeError):
        se(np.int32(1), default=None, option=None)

    with pytest.raises(TypeError):
        se(np.array([1, 2, 3]), default=None, option=None)

    @serde.serde
    class MisTypedNoDefaultEncoder:
        a: int
        b: float
        h: bool
        c: list[int]
        d: list[float]
        e: list[bool]

    def unchecked_new(
        a: np.int32,
        b: np.float32,
        h: np.bool_,
        c: npt.NDArray[np.int32],
        d: npt.NDArray[np.float32],
        e: npt.NDArray[np.bool_],
    ) -> MisTypedNoDefaultEncoder:
        obj = MisTypedNoDefaultEncoder(
            0,
            0.0,
            True,
            [],
            [],
            [],
        )
        obj.a = a  # type: ignore
        obj.b = b  # type: ignore
        obj.h = h  # type: ignore
        obj.c = c  # type: ignore
        obj.d = d  # type: ignore
        obj.e = e  # type: ignore
        return obj

    test1 = unchecked_new(
        np.int32(1),
        np.float32(3.0),
        np.bool_(False),
        np.array([np.int32(i) for i in [1, 2]]),
        np.array([np.float32(i) for i in [5.0, 6.0]]),
        np.array([np.bool_(i) for i in [True, False]]),
    )

    with pytest.raises(TypeError):
        se(test1, default=None, option=None)

    test2 = unchecked_new(
        np.int64(1),
        np.float64(3.0),
        np.bool_(False),
        np.array([np.int64(i) for i in [1, 2]]),
        np.array([np.float64(i) for i in [5.0, 6.0]]),
        np.array([np.bool_(i) for i in [True, False]]),
    )

    with pytest.raises(TypeError):
        se(test2, default=None, option=None)


def test_numpy_misc():
    assert serde.numpy.fullname(str) == "str"



import logging
from dataclasses import dataclass
from typing import Any, Literal

import pytest

import serde

from .common import (
    all_formats,
    format_dict,
    format_json,
    format_msgpack,
    format_tuple,
    format_yaml,
    opt_case,
    opt_case_ids,
)

log = logging.getLogger("test")

serde.init(True)


@serde.serde
@dataclass(unsafe_hash=True)
class LitInt:
    """
    Integer.
    """

    i: Literal[1, 2]


@serde.serde
@dataclass(unsafe_hash=True)
class LitStr:
    """
    String.
    """

    s: Literal["foo", "bar"]


@serde.serde
@dataclass(unsafe_hash=True)
class LitBool:
    """
    Boolean.
    """

    b: Literal[True]


@serde.serde
@dataclass(unsafe_hash=True)
class LitMixed:
    """
    Mixed Literal
    """

    m: Literal[1, 2, "foo", "bar", False, True]


@serde.serde
@dataclass(unsafe_hash=True)
class LitDict:
    """
    Dict containing primitives.
    """

    i: dict[Literal[1, 2], Literal[1, 2]]
    s: dict[Literal["foo", "bar"], Literal["foo", "bar"]]
    b: dict[Literal[True], Literal[True]]
    m: dict[
        Literal[1, 2, "foo", "bar", False, True],
        Literal[1, 2, "foo", "bar", False, True],
    ]


@serde.serde
@dataclass(unsafe_hash=True)
class LitStrDict:
    """
    Dict containing primitives.
    """

    i: dict[Literal["1", "2"], Literal[1, 2]]
    s: dict[Literal["foo", "bar"], Literal["foo", "bar"]]
    b: dict[Literal["True"], Literal[True]]
    m: dict[
        Literal["1", "2", "foo", "bar", "False", "True"],
        Literal[1, 2, "foo", "bar", False, True],
    ]


@serde.serde
@dataclass(unsafe_hash=True)
class Literals:
    """
    Primitives.
    """

    i: Literal[1, 2]
    s: Literal["foo", "bar"]
    b: Literal[True]
    m: Literal[1, 2, "foo", "bar", False, True]


@serde.serde
class LitNestedPrituple:
    """
    tuple containing nested primitives.
    """

    i: tuple[LitInt, LitInt]
    s: tuple[LitStr, LitStr]
    b: tuple[LitBool, LitBool, LitBool]
    m: tuple[LitMixed, LitMixed]


listLiterals = list[Literals]
PRI = Literals(1, "foo", True, 2)
DictLiterals = dict[str, Literals]


"""
Tests
"""


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_dict(se: Any, de: Any, opt: Any) -> None:
    if se in (serde.json.to_json, serde.msgpack.to_msgpack, serde.toml.to_toml):
        # JSON, Msgpack, Toml don't allow non string key.
        p = LitStrDict({"1": 2}, {"foo": "bar"}, {"True": True}, {"foo": True})
        assert p == de(LitStrDict, se(p))
    else:
        p = LitDict({1: 2}, {"foo": "bar"}, {True: True}, {2: False})  # type: ignore[assignment]
        assert p == de(LitDict, se(p))


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_tuple(se: Any, de: Any, opt: Any) -> None:
    if se is not serde.toml.to_toml:
        p = LitNestedPrituple(
            (LitInt(1), LitInt(2)),
            (LitStr("foo"), LitStr("bar")),
            (LitBool(True), LitBool(True), LitBool(True)),
            (LitMixed(False), LitMixed("foo")),
        )
        assert p == de(LitNestedPrituple, se(p))


@pytest.mark.parametrize(
    "se,de", (format_dict + format_tuple + format_json + format_msgpack + format_yaml)
)
def test_list_literals(se: Any, de: Any) -> None:
    p = [PRI, PRI]
    assert p == de(listLiterals, se(p))

    p = []
    assert p == de(listLiterals, se(p))


@pytest.mark.parametrize(
    "se,de", (format_dict + format_tuple + format_json + format_msgpack + format_yaml)
)
def test_dict_literals(se: Any, de: Any) -> None:
    p = {"1": PRI, "2": PRI}
    assert p == de(DictLiterals, se(p))

    p = {}
    assert p == de(DictLiterals, se(p))


def test_json() -> None:
    p = Literals(1, "foo", True, "bar")
    s = '{"i":1,"s":"foo","b":true,"m":"bar"}'
    assert s == serde.json.to_json(p)


def test_msgpack() -> None:
    p = Literals(1, "foo", True, "bar")
    d = b"\x84\xa1i\x01\xa1s\xa3foo\xa1b\xc3\xa1m\xa3bar"
    assert d == serde.msgpack.to_msgpack(p)
    assert p == serde.msgpack.from_msgpack(Literals, d)


def test_msgpack_unnamed() -> None:
    p = Literals(1, "foo", True, "bar")
    d = b"\x94\x01\xa3foo\xc3\xa3bar"
    assert d == serde.msgpack.to_msgpack(p, named=False)
    assert p == serde.msgpack.from_msgpack(Literals, d, named=False)



from dataclasses import dataclass, field
from functools import partial
from typing import Union

import data
from mashumaro.mixins.json import DataClassJSONMixin
from runner import Runner, Size


@dataclass
class Small(DataClassJSONMixin):
    i: int
    s: str
    f: float
    b: bool


@dataclass
class Medium(DataClassJSONMixin):
    inner: list[Small] = field(default_factory=list)


@dataclass
class Large(DataClassJSONMixin):
    customer_id: int
    name: str
    email: str
    preferences: dict[str, Union[str, bool, int]] = field(default_factory=dict)
    items_list: list[str] = field(default_factory=list)
    nested_data: dict[str, list[int]] = field(default_factory=dict)
    loyalty_points: int = 0
    created_at: str = ""


SMALL = Small(**data.args_sm)

MEDIUM = Medium([Small(**d) for d in data.args_md])


# Create Large instance
def create_large_instance() -> Large:
    return Large(
        customer_id=12345,
        name="John Smith",
        email="john@example.com",
        preferences={
            "theme": "dark",
            "notifications": True,
            "language": "en",
            "max_budget": 5000,
            "auto_renew": False,
            "privacy_level": 3,
        },
        items_list=["laptop", "mouse", "keyboard", "monitor", "speakers"] * 20,  # 100 items
        nested_data={
            "category_1": list(range(50)),
            "category_2": list(range(50, 100)),
            "category_3": list(range(100, 150)),
            "category_4": list(range(150, 200)),
            "category_5": list(range(200, 250)),
        },
        loyalty_points=1250,
        created_at="2024-01-15T10:30:00Z",
    )


LARGE = create_large_instance()


def new(size: Size) -> Runner:
    name = "mashumaro"
    if size == Size.Small:
        unp = SMALL
        pac = data.SMALL
        cls = Small
    elif size == Size.Medium:
        unp = MEDIUM  # type: ignore[assignment]
        pac = data.MEDIUM
        cls = Medium  # type: ignore[assignment]
    elif size == Size.Large:
        unp = LARGE  # type: ignore[assignment]
        pac = data.LARGE
        cls = Large  # type: ignore[assignment]
    return Runner(name, unp, partial(se, unp), partial(de, cls, pac), None, None)


def se(obj: Union[Small, Medium, Large]) -> str:
    return obj.to_json()  # type: ignore[return-value]


def de(
    cls: Union[type[Small], type[Medium], type[Large]], data: str
) -> Union[Small, Medium, Large]:
    return cls.from_json(data)



import sys
import pytest
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Generic, NewType, Optional, TypeVar, Union, Literal

from serde import serde, field, is_serializable, is_deserializable, SerdeError
from serde.compat import (
    get_generic_arg,
    is_dict,
    is_generic,
    is_list,
    is_opt,
    is_primitive,
    is_set,
    is_tuple,
    is_union,
    iter_types,
    iter_unions,
    type_args,
    typename,
    union_args,
)
from serde.core import is_instance

from .data import Bool, Float, Int, Pri, PriOpt, Str

T = TypeVar("T")
U = TypeVar("U")


def test_types() -> None:
    assert is_list(list[int])
    assert is_list(list)
    assert is_set(set[int])
    assert is_set(set)
    assert is_tuple(tuple[int, int, int])
    assert is_tuple(tuple)
    assert is_tuple(tuple[int, ...])
    assert is_dict(dict[str, int])
    assert is_dict(dict)
    assert is_opt(Optional[int])
    assert is_opt(Union[int, None])
    assert is_union(Union[int, str])
    assert is_union(Union[Optional[int], Optional[str]])
    assert is_union(Optional[int])
    assert not is_opt(Union[Optional[int], Optional[str]])
    assert is_union(Union[Optional[int], Optional[str]])

    assert is_list(list)
    assert is_set(set)
    assert is_tuple(tuple)
    assert is_dict(dict)

    Int = NewType("Int", int)
    assert is_primitive(int)
    assert is_primitive(Int)  # type: ignore[arg-type]

    if sys.version_info[:3] >= (3, 10, 0):
        assert is_union(str | int)
        assert is_union(str | None)
        assert is_opt(str | None)


def test_typename() -> None:
    @serde
    class Bar(Generic[T]):
        v: T

    @serde
    class Foo(Generic[T]):
        nested: Bar[T]

    assert typename(Foo[int]) == "Foo"
    assert typename(Foo) == "Foo"
    assert typename(list[int]) == "list[int]"
    assert typename(Optional) == "Optional"
    assert typename(list) == "list"
    assert typename(list[int]) == "list[int]"
    assert typename(tuple) == "tuple"
    assert typename(tuple[int, str]) == "tuple[int, str]"
    assert typename(tuple[int, ...]) == "tuple[int, ...]"
    assert typename(dict) == "dict"
    assert typename(dict[str, Foo]) == "dict[str, Foo]"  # type: ignore
    assert typename(set) == "set"
    assert typename(set[int]) == "set[int]"
    assert typename(Literal[1, True, "Hey"]) == "Literal[1, True, 'Hey']"


def test_iter_types() -> None:
    assert {Pri, int, str, float, bool} == set(iter_types(Pri))
    assert {dict, str, Pri, int, float, bool} == set(iter_types(dict[str, Pri]))
    assert {list, str} == set(iter_types(list[str]))
    assert {set, str} == set(iter_types(set[str]))
    assert {tuple, int, str, bool, float} == set(iter_types(tuple[int, str, bool, float]))
    assert {tuple, int, Ellipsis} == set(iter_types(tuple[int, ...]))
    assert {PriOpt, Optional, int, str, float, bool} == set(iter_types(PriOpt))

    @serde
    class Foo:
        a: int
        b: datetime
        c: datetime
        d: Optional[str] = None
        e: Union[str, int] = 10
        f: list[int] = field(default_factory=list)
        g: set[int] = field(default_factory=set)

    assert {Foo, datetime, Optional, str, Union, list, set, int} == set(iter_types(Foo))


def test_iter_unions() -> None:
    assert [Union[str, int]] == list(iter_unions(Union[str, int]))
    assert [Union[str, int]] == list(iter_unions(dict[str, Union[str, int]]))
    assert [Union[str, int]] == list(iter_unions(tuple[Union[str, int]]))
    assert [Union[str, int]] == list(iter_unions(list[Union[str, int]]))
    assert [Union[str, int]] == list(iter_unions(set[Union[str, int]]))
    assert [Union[str, int, type(None)]] == list(iter_unions(Optional[Union[str, int]]))

    @dataclass
    class A:
        a: list[Union[int, str]]
        b: dict[str, list[Union[float, int]]]
        C: dict[Union[bool, str], Union[float, int]]

    assert {Union[int, str], Union[float, int], Union[bool, str], Union[float, int]} == set(
        iter_unions(A)
    )


def test_type_args() -> None:
    assert (int,) == type_args(list[int])
    assert (int, str) == type_args(dict[int, str])
    assert (int, type(None)) == type_args(Optional[int])
    assert (Optional[int],) == type_args(list[Optional[int]])  # type: ignore
    assert (list[int], type(None)) == type_args(Optional[list[int]])
    assert (list[int], dict[str, int]) == type_args(Union[list[int], dict[str, int]])
    assert (int, type(None), str) == type_args(Union[Optional[int], str])
    assert (int, Ellipsis) == type_args(tuple[int, ...])  # type: ignore


def test_union_args() -> None:
    assert (int, str) == union_args(Union[int, str])
    assert (list[int], dict[int, str]) == union_args(Union[list[int], dict[int, str]])
    assert (Optional[int], str) == union_args(Union[Optional[int], str])  # type: ignore


def test_is_instance() -> None:
    # primitive
    assert is_instance(10, int)
    assert is_instance("str", str)
    assert is_instance(1.0, float)
    assert is_instance(False, bool)
    assert not is_instance(10.0, int)
    assert not is_instance("10", int)
    assert is_instance(True, int)  # see why this is true https://stackoverflow.com/a/37888668

    # dataclass
    p = Pri(i=10, s="foo", f=100.0, b=True)
    assert is_instance(p, Pri)
    p.i = 10.0  # type: ignore
    assert is_instance(p, Pri)  # there is no way to check mulated properties.

    # dataclass (Nested)
    @dataclass
    class Foo:
        p: Pri

    h = Foo(Pri(i=10, s="foo", f=100.0, b=True))
    assert is_instance(h, Foo)
    h.p.i = 10.0  # type: ignore
    assert is_instance(h, Foo)

    # list
    assert is_instance([], list[int])
    assert is_instance([10], list)
    assert is_instance([10], list)
    assert is_instance([10], list[int])
    assert not is_instance([10.0], list[int])

    # list of dataclass
    assert is_instance([Int(n) for n in range(1, 10)], list[Int])
    assert not is_instance([Str("foo")], list[Int])

    # set
    assert is_instance(set(), set[int])
    assert is_instance({10}, set)
    assert is_instance({10}, set[int])
    assert not is_instance({10.0}, set[int])

    # set of dataclass
    assert is_instance({Int(n) for n in range(1, 10)}, set[Int])
    assert not is_instance({Str("foo")}, set[Int])
    # tuple
    assert not is_instance((), tuple[int, str, float, bool])
    assert is_instance((10, "a"), tuple)
    assert is_instance((10, "foo", 100.0, True), tuple[int, str, float, bool])
    assert not is_instance((10, "foo", 100.0, "last-type-is-wrong"), tuple[int, str, float, bool])
    assert is_instance((10, 20), tuple[int, ...])
    assert is_instance((10, 20, 30), tuple[int, ...])
    assert is_instance((), tuple[()])
    assert is_instance((), tuple[int, ...])
    assert not is_instance((10, "a"), tuple[int, ...])

    # tuple of dataclasses
    assert is_instance(
        (Int(10), Str("foo"), Float(100.0), Bool(True)), tuple[Int, Str, Float, Bool]
    )
    assert not is_instance(
        (Int(10), Str("foo"), Str("wrong-class"), Bool(True)), tuple[Int, Str, Float, Bool]
    )

    # dict
    assert is_instance({}, dict[str, int])
    assert is_instance({"a": "b"}, dict)
    assert is_instance({"a": "b"}, dict)
    assert is_instance({"foo": 10, "bar": 20}, dict[str, int])
    assert not is_instance({"foo": 10.0, "bar": 20}, dict[str, int])

    # dict of dataclass
    assert is_instance({Str("foo"): Int(10), Str("bar"): Int(20)}, dict[Str, Int])
    assert not is_instance({Str("foo"): Str("wrong-type"), Str("bar"): Int(10)}, dict[Str, Int])

    # optional
    assert is_instance(None, type(None))
    assert is_instance(10, Optional[int])
    assert is_instance(None, Optional[int])
    assert not is_instance("wrong-type", Optional[int])

    # optional of dataclass
    assert is_instance(Int(10), Optional[Int])
    assert not is_instance("wrong-type", Optional[Int])

    # nested containers
    assert is_instance([({"a": "b"}, 10, [True])], list[tuple[dict[str, str], int, list[bool]]])
    assert not is_instance(
        [({"a": "b"}, 10, ["wrong-type"])], list[tuple[dict[str, str], int, list[bool]]]
    )


@serde
class GenericFoo(Generic[T]):
    t: T


def test_is_generic() -> None:
    assert not is_generic(int)
    assert not is_generic(list[int])
    assert not is_generic(list)
    assert not is_generic(GenericFoo)
    assert is_generic(GenericFoo[int])
    assert is_generic(GenericFoo[list[int]])
    assert not is_generic(Optional[int])
    assert not is_generic(Optional[list[int]])
    assert is_serializable(GenericFoo)
    assert not is_serializable(GenericFoo[list[int]])
    assert is_deserializable(GenericFoo)
    assert not is_deserializable(GenericFoo[list[int]])


def test_get_generic_arg() -> None:
    class GenericFoo(Generic[T, U]):
        pass

    assert get_generic_arg(GenericFoo[int, str], ["T", "U"], ["T", "U"], 0) is int
    assert get_generic_arg(GenericFoo[int, str], ["T", "U"], ["T", "U"], 1) is str
    assert get_generic_arg(GenericFoo[int, str], ["T", "U"], ["U"], 0) is str
    assert get_generic_arg(GenericFoo[int, str], ["T", "U"], ["V"], 0) == Any

    with pytest.raises(SerdeError):
        get_generic_arg(GenericFoo[int, str], ["T"], ["T"], 0)

    with pytest.raises(SerdeError):
        get_generic_arg(GenericFoo[int, str], ["T"], ["U"], 0)



from serde import serde
from serde.toml import to_toml, from_toml
from typing import Optional
import pytest


def test_toml_basics() -> None:
    @serde
    class Foo:
        v: Optional[int]

    f = Foo(10)
    assert "v = 10\n" == to_toml(f)
    assert f == from_toml(Foo, "v = 10\n")

    @serde
    class Bar:
        v: set[int]

    toml_literal = """\
v = [
    1,
    2,
    3,
]
"""
    b = Bar({1, 2, 3})
    assert toml_literal == to_toml(b)
    assert b == from_toml(Bar, toml_literal)


def test_skip_none() -> None:
    @serde
    class Foo:
        a: int
        b: Optional[int]

    f = Foo(10, 100)
    assert (
        to_toml(f)
        == """\
a = 10
b = 100
"""
    )

    f = Foo(10, None)
    assert (
        to_toml(f)
        == """\
a = 10
"""
    )


def test_skip_none_container_not_supported_yet() -> None:
    @serde
    class Foo:
        a: int
        b: list[Optional[int]]

    f = Foo(10, [100, None])
    with pytest.raises(TypeError):
        to_toml(f)



# type: ignore  # SQLAlchemy types imported via pytest.importorskip
import logging
from typing import Optional, Any

import pytest

import serde
from .common import (
    opt_case,
    opt_case_ids,
    FormatFuncs,
    format_dict,
    format_tuple,
    format_json,
    format_msgpack,
    format_yaml,
    format_pickle,
    format_toml,
    all_formats,
)

sa = pytest.importorskip("sqlalchemy", "2.0.0")
sa_orm = pytest.importorskip("sqlalchemy.orm")

log = logging.getLogger("test")

serde.init(True)

all_except_toml: FormatFuncs = (
    format_dict + format_tuple + format_json + format_msgpack + format_yaml + format_pickle
)


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_except_toml)
def test_sqlalchemy_simple(se: Any, de: Any, opt: Any) -> None:
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    class Base(sa_orm.MappedAsDataclass, sa_orm.DeclarativeBase):
        pass

    @serde.serde(**opt)
    class User(Base):
        __tablename__ = "user"

        id: sa_orm.Mapped[int] = sa_orm.mapped_column(primary_key=True)
        name: sa_orm.Mapped[str]
        fullname: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String(30))
        nickname: sa_orm.Mapped[Optional[str]] = sa_orm.mapped_column(init=False)

    user = User(1, "john", "John Doe")
    assert user == de(User, se(user))


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", format_toml)
def test_sqlalchemy_simple_toml(se: Any, de: Any, opt: Any) -> None:
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    class Base(sa_orm.MappedAsDataclass, sa_orm.DeclarativeBase):
        pass

    @serde.serde(**opt)
    class User(Base):
        __tablename__ = "user"

        id: sa_orm.Mapped[int] = sa_orm.mapped_column(primary_key=True)
        name: sa_orm.Mapped[str]
        fullname: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.String(30))

    user = User(1, "john", "John Doe")
    assert user == de(User, se(user))


@pytest.mark.parametrize("opt", opt_case, ids=opt_case_ids())
@pytest.mark.parametrize("se,de", all_formats)
def test_sqlalchemy_nested(se: Any, de: Any, opt: Any) -> None:
    log.info(f"Running test with se={se.__name__} de={de.__name__} opts={opt}")

    class Base(sa_orm.MappedAsDataclass, sa_orm.DeclarativeBase):
        pass

    @serde.serde(**opt)
    class Project(Base):
        __tablename__ = "projects"
        id: sa_orm.Mapped[int] = sa_orm.mapped_column(primary_key=True)
        name: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.Text, nullable=False)
        owner_id: sa_orm.Mapped[int] = sa_orm.mapped_column(sa.ForeignKey("users.id"))

    @serde.serde(**opt)
    class User(Base):
        __tablename__ = "users"

        id: sa_orm.Mapped[int] = sa_orm.mapped_column(primary_key=True)
        name: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.Text, nullable=False)
        fullname: sa_orm.Mapped[str] = sa_orm.mapped_column(sa.Text, nullable=False)
        projects: sa_orm.Mapped[list[Project]] = sa_orm.relationship(backref="owner")

    user = User(1, "john", "John Doe", [Project(1, "Dummy", 1)])
    assert user == de(User, se(user))



from __future__ import annotations
from dataclasses import dataclass

from serde import serde
from serde.json import from_json, to_json


@dataclass
class Node:
    name: str
    children: list[Node]


serde(Node)


def main() -> None:
    n = Node("a", [Node("b", [Node("c", [])])])
    s = to_json(n)
    print(f"Into Json: {s}")
    print(f"From Json: {from_json(Node, s)}")


if __name__ == "__main__":
    main()



from dataclasses import InitVar

from serde import serde, field
from serde.json import from_json, to_json


@serde
class Foo:
    a: int
    b: int | None = field(default=None, init=False)
    c: InitVar[int | None] = 1000

    def __post_init__(self, c: int | None) -> None:
        self.b = self.a * 10


def main() -> None:
    f = Foo(10)
    print(f"Into Json: {to_json(f)}")

    s = '{"a": 10}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == "__main__":
    main()



from serde import serde, Untagged, AdjacentTagging, InternalTagging
from serde.json import from_json, to_json


@serde
class Bar:
    b: int


@serde
class Baz:
    b: int


def main() -> None:
    baz = Baz(10)

    print("# external tagging (default)")
    a = to_json(baz, cls=Bar | Baz)
    print("IntoJSON", a)
    print("FromJSON", from_json(Bar | Baz, a))

    print("# internal tagging")
    a = to_json(baz, cls=InternalTagging("type", Bar | Baz))
    print("IntoJSON", a)
    print("FromJSON", from_json(InternalTagging("type", Bar | Baz), a))

    print("# adjacent tagging")
    a = to_json(baz, cls=AdjacentTagging("type", "content", Bar | Baz))
    print("IntoJSON", a)
    print("FromJSON", from_json(AdjacentTagging("type", "content", Bar | Baz), a))

    print("# untagged")
    a = to_json(baz, cls=Untagged(Bar | Baz))
    print("IntoJSON", a)
    print("FromJSON", from_json(Untagged(Bar | Baz), a))


if __name__ == "__main__":
    main()



import json
from functools import partial
from typing import Any, Union

import cattrs
import data
from attrs_class import LARGE, MEDIUM, SMALL, Large, Medium, Small
from runner import Runner, Size


def new(size: Size) -> Runner:
    name = "cattrs"
    if size == Size.Small:
        unp = SMALL
        pac = data.SMALL
        cls = Small
    elif size == Size.Medium:
        unp = MEDIUM
        pac = data.MEDIUM
        cls = Medium
    elif size == Size.Large:
        unp = LARGE
        pac = data.LARGE
        cls = Large
        # Skip deserialization for Large due to Union type issues
        return Runner(name, unp, partial(se, unp), None, None, partial(asdict, unp))
    return Runner(name, unp, partial(se, unp), partial(de, cls, pac), None, partial(asdict, unp))


def se(obj: Union[Small, Medium, Large]) -> str:
    return json.dumps(cattrs.unstructure(obj))


def de(cls: type, data: str) -> Union[Small, Medium, Large]:
    return cattrs.structure(json.loads(data), cls)


def asdict(obj: Union[Small, Medium, Large]) -> dict[str, Any]:
    return cattrs.unstructure(obj)  # type: ignore[no-any-return]
